import collections
import hashlib
from pathlib import Path
import subprocess

from suitcase.utils import SuitcaseUtilsTypeError, SuitcaseUtilsValueError, ModeError


class SSHManager:
    """
    Write to files on a remote machine via ssh <host> 'cat <path>'

    This is really gross and the author absolves knowlege of it.
    """

    def __init__(self, host, directory):
        self._reserved_names = set()
        self._artifacts = collections.defaultdict(list)
        self._host = host
        self._directory = Path(directory)
        self.buffers = {}  # maps postfixes to buffer objects

    def __enter__(self):
        return self

    def __exit__(self, *exc_details):
        self.close()

    @property
    def artifacts(self):
        return dict(self._artifacts)

    def reserve_name(self, label, postfix):
        """
        This action is not valid on this manager. It will always raise.
        Parameters
        ----------
        label : string
            A label for the sort of content being stored, such as
            'stream_data' or 'metadata'.
        postfix : string
            Relative file path. Must be unique for this Manager.
        Raises
        ------
        SuitcaseUtilsTypeError
        """
        raise SuitcaseUtilsTypeError(
            f"{type(self)} is incompatible with exporters that require "
            f"explicit filenames."
        )

    def open(self, label, postfix, mode, encoding=None, errors=None):
        """
        Request a file handle.

        Like the built-in open function, this may be used as a context manager.

        Parameters
        ----------
        label : string
            A label for the sort of content being stored, such as
            'stream_data' or 'metadata'.
        postfix : string
            Relative file path (simply used as an identifer in this case, as
            there is no actual file). Must be unique for this Manager.
        mode : {'x', 'xt', xb'}
            'x' or 'xt' for text, 'xb' for binary
        encoding : string or None
            Not used. Accepted for compatibility with built-in open().
        errors : string or None
            Not used. Accepted for compatibility with built-in open().
        Returns
        -------
        file : handle
        """
        # Of course, in-memory buffers have no filepath, but we still expect
        # postfix to be a thing that looks like a relative filepath, and we use
        # it as a unique identifier for a given buffer.
        if Path(postfix).is_absolute():
            raise SuitcaseUtilsValueError(
                f"The postfix {postfix} must be structured like a relative "
                f"file path."
            )
        name = Path(postfix).expanduser()
        if name in self._reserved_names:
            raise SuitcaseUtilsValueError(
                f"The postfix {postfix!r} has already been used."
            )
        self._reserved_names.add(name)
        if mode in ("x", "xt"):
            buffer = PipeStringToCat(self._host, str(self._directory / name))
        elif mode == "xb":
            buffer = PipeBytesToCat(self._host, str(self._directory / name))
        else:
            raise ModeError(
                f"The mode passed to MemoryBuffersManager.open is {mode} but "
                f"needs to be one of 'x', 'xt' or 'xb'."
            )
        self._artifacts[label].append(buffer)
        self.buffers[postfix] = buffer
        return buffer

    def close(self):
        """Close all buffers opened by the manager.
        """
        for f in self.buffers.values():
            f.close()


class PipeToCat:
    "A base class for piping to ssh <host> 'cat <path>'"

    def __init__(self, host, path):
        self._host = host
        self._path = path
        # Ensure directory exists.
        subprocess.run(
            ["ssh", host, "mkdir", "-p", str(Path(path).parent)], capture_output=True
        )
        self._process = subprocess.Popen(
            ["ssh", host, f"cat > {path}"],
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )

    def __repr__(self):
        return f"{self.__class__.__name__}({self._host}, {self._path})"

    @property
    def path(self):
        return self._path

    def close(self):
        self._process.stdin.close()
        self._process.wait()

    def __enter__(self):
        return self

    def __exit__(self, *exc_details):
        self.close()


class PipeStringToCat(PipeToCat):
    def write(self, s):
        return self._process.stdin.write(s.encode())


class PipeBytesToCat(PipeToCat):
    def write(self, b):
        return self._process.stdin.write(b)


def root_hash(salt, root):
    """
    Generate a deterministic hash for a given salt and root.

    Parameters
    ----------
    salt: bytes
    root: Union[string, Path]

    Examples
    --------

    Generate a salt and hash a root (string or Path).

    >>> import secrets
    >>> salt = secrets.token_hex(32).encode()
    >>> root_hash(salt, root)
    """
    return hashlib.md5(str(root).encode() + salt).hexdigest()


class CatalogNameExists(ValueError):
    ...



import collections
import functools
import logging
import os
import pathlib
import secrets
import shutil

import event_model
import databroker.core
from tqdm import tqdm
import yaml

from ._utils import root_hash
from ._version import get_versions

__all__ = (
    "copy_external_files",
    "export_uids",
    "export_catalog",
    "export_run",
    "write_documents_manifest",
    "write_external_files_manifest",
    "write_jsonl_catalog_file",
    "write_msgpack_catalog_file",
)
logger = logging.getLogger(__name__)


# Write through tqdm to avoid overlapping with bars.
def print(*args):
    tqdm.write(" ".join(str(arg) for arg in args))


def export_uids(
    source_catalog,
    uids,
    directory,
    *,
    strict=False,
    external=None,
    no_documents=False,
    handler_registry=None,
    serializer_class=None,
    salt=None,
):
    """
    Export Runs from a Catalog, given a list of RunStart unique IDs.

    Parameters
    ----------
    source_catalog: Catalog
    uids: List[Str]
        List of RunStart unique IDs
    directory: Union[Str, Manager]
        Where files containing documents will be written, or a Manager for
        writing to non-file buffers.
    strict: Bool, optional
        By default, swallow erros and return a list of them at the end.
        Set to True to debug errors.
    external: {None, 'fill', 'ignore')
        If None, return the paths to external files.
        If 'fill', fill the external data into the Documents.
        If 'ignore', do not locate external files.
    no_documents: Bool, optional
        If True, do not write any files. False by default.
    handler_registry: Union[Dict, None]
        If None, automatic handler discovery is used.
    serializer_class: Serializer
        Expected to be a lossless serializer that encodes a format for which
        there is a corresponding databroker intake driver. Default (None) is
        currently ``suitcase.msgpack.Serializer``, but this may change in the
        future. If you want ``suitcase.msgpack.Serializer`` specifically, pass
        it in explicitly.
    salt: Union[bytes, None]
        We want to make hashes is unique to:

        - a root
        - a given batch of exported runs (i.e., a given call to this function)

        so that we can use it as a key in root_map which is guaranteed not to
        collide with keys from other batches. Thus, we create a "salt" unless
        one is specified here. This does not need to be cryptographically
        secure, just unique.

    Returns
    -------
    artifacts, files, failures

    Notes
    -----
    * ``artifacts`` maps a human-readable string (typically just ``'all'`` in
      this case) to a list of buffers or filepaths where the documents were
      serialized.
    * ``files`` is the set of filepaths of all external files referenced by
      Resource documents, keyed on ``(root_in_document, root, unique_id)``.
    * ``failures`` is a list of uids of runs that raised Exceptions. (The
      relevant tracebacks are logged.)
    * ``file_uids`` is a dictionary of RunStart unique IDs mapped to a set of
      ``(root, filename)`` pairs.
    """
    accumulated_files = collections.defaultdict(set)
    file_uids = {}
    accumulated_artifacts = collections.defaultdict(set)
    failures = []
    if salt is None:
        salt = secrets.token_hex(32).encode()
    root_hash_func = functools.partial(root_hash, salt)
    with tqdm(total=len(uids), position=1, desc="Writing Documents") as progress:
        for uid in uids:
            try:
                run = source_catalog[uid]
                artifacts, files = export_run(
                    run,
                    directory,
                    root_hash_func,
                    external=external,
                    no_documents=no_documents,
                    handler_registry=handler_registry,
                    root_map=source_catalog.root_map,
                    serializer_class=serializer_class,
                )
                file_uids[uid] = []
                for root, set_ in files.items():
                    accumulated_files[root].update(set_)
                    file_uids[uid].extend((root, filename) for filename in set_)
                for name, list_ in artifacts.items():
                    accumulated_artifacts[name].update(list_)

            except Exception:
                logger.exception("Error while exporting Run %r", uid)
                if strict:
                    raise
                failures.append(uid)
                progress.set_description(
                    f"Writing Documents ({len(failures)} failures)", refresh=False
                )
            progress.update()
    return dict(accumulated_artifacts), dict(accumulated_files), failures, file_uids


def export_catalog(
    source_catalog,
    directory,
    *,
    strict=False,
    external=None,
    no_documents=False,
    handler_registry=None,
    serializer_class=None,
    salt=None,
    limit=None,
):
    """
    Export all the Runs from a Catalog.

    Parameters
    ----------
    source_catalog: Catalog
    directory: Union[Str, Manager]
        Where files containing documents will be written, or a Manager for
        writing to non-file buffers.
    strict: Bool, optional
        By default, swallow erros and return a list of them at the end.
        Set to True to debug errors.
    external: {None, 'fill', 'ignore')
        If None, return the paths to external files.
        If 'fill', fill the external data into the Documents.
        If 'ignore', do not locate external files.
    no_documents: Bool, optional
        If True, do not serialize documents. False by default.
    handler_registry: Union[Dict, None]
        If None, automatic handler discovery is used.
    serializer_class: Serializer
        Expected to be a lossless serializer that encodes a format for which
        there is a corresponding databroker intake driver. Default (None) is
        currently ``suitcase.msgpack.Serializer``, but this may change in the
        future. If you want ``suitcase.msgpack.Serializer`` specifically, pass
        it in explicitly.
    salt: Union[bytes, None]
        We want to make hashes is unique to:

        - a root
        - a given batch of exported runs (i.e., a given call to this function)

        so that we can use it as a key in root_map which is guaranteed not to
        collide with keys from other batches. Thus, we create a "salt" unless
        one is specified here. This does not need to be cryptographically
        secure, just unique.
    limit: Union[Integer, None]
        Stop after exporting some number of Runs. Useful for testing a subset
        before doing a lengthy export.

    Returns
    -------
    artifacts, files, failures

    Notes
    -----
    * ``artifacts`` maps a human-readable string (typically just ``'all'`` in
      this case) to a list of buffers or filepaths where the documents were
      serialized.
    * ``files`` is the set of filepaths of all external files referenced by
      Resource documents, keyed on ``(root_in_document, root, unique_id)``.
    * ``failures`` is a list of uids of runs that raised Exceptions. (The
      relevant tracebacks are logged.)
    * ``file_uids`` is a dictionary of RunStart unique IDs mapped to a set of
      ``(root, filename)`` pairs.
    """
    if limit is not None:
        if limit < 1:
            raise ValueError("limit must be None or a number 1 or greater")
        limit = int(limit)
    accumulated_files = collections.defaultdict(set)
    file_uids = {}
    accumulated_artifacts = collections.defaultdict(set)
    failures = []
    if salt is None:
        salt = secrets.token_hex(32).encode()
    root_hash_func = functools.partial(root_hash, salt)
    with tqdm(
        total=limit or len(source_catalog), position=1, desc="Writing Documents"
    ) as progress:
        for i, (uid, run) in enumerate(source_catalog.items()):
            if i == limit:
                break
            try:
                artifacts, files = export_run(
                    run,
                    directory,
                    root_hash_func,
                    external=external,
                    no_documents=no_documents,
                    handler_registry=handler_registry,
                    root_map=source_catalog.root_map,
                    serializer_class=serializer_class,
                )
                file_uids[uid] = []
                for root, set_ in files.items():
                    accumulated_files[root].update(set_)
                    file_uids[uid].extend((root, filename) for filename in set_)
                for name, list_ in artifacts.items():
                    accumulated_artifacts[name].update(list_)
            except Exception:
                logger.exception("Error while exporting Run %r", uid)
                if strict:
                    raise
                failures.append(uid)
                progress.set_description(
                    f"Writing Documents ({len(failures)} failures)", refresh=False
                )
            progress.update()
    return dict(accumulated_artifacts), dict(accumulated_files), failures, file_uids


def export_run(
    run,
    directory,
    root_hash_func,
    *,
    external=None,
    no_documents=False,
    handler_registry=None,
    root_map=None,
    serializer_class=None,
):
    """
    Export one Run.

    Parameters
    ----------
    run: BlueskyRun
    directory: Union[Str, Manager]
        Where files containing documents will be written, or a Manager for
        writing to non-file buffers.
    external: {None, 'fill', 'ignore')
        If None, return the paths to external files.
        If 'fill', fill the external data into the Documents.
        If 'ignore', do not locate external files.
    no_documents: Bool, optional
        If True, do not serialize documents. False by default.
    handler_registry: Union[Dict, None]
        If None, automatic handler discovery is used.
    serializer_class: Serializer, optional
        Expected to be a lossless serializer that encodes a format for which
        there is a corresponding databroker intake driver. Default (None) is
        currently ``suitcase.msgpack.Serializer``, but this may change in the
        future. If you want ``suitcase.msgpack.Serializer`` specifically, pass
        it in explicitly.

    Returns
    -------
    artifacts, files

    Notes
    -----
    * ``artifacts`` maps a human-readable string (typically just ``'all'`` in
      this case) to a list of buffers or filepaths where the documents were
      serialized.
    * ``files`` is the set of filepaths of all external files referenced by
      Resource documents, keyed on ``(root_in_document, root, unique_id)``.
    """
    EXTERNAL_RELATED_DOCS = ("resource", "datum", "datum_page")
    if serializer_class is None:
        import suitcase.msgpack

        serializer_class = suitcase.msgpack.Serializer
    root_map = root_map or {}
    files = collections.defaultdict(set)
    if handler_registry is None:
        handler_registry = databroker.core.discover_handlers()
    with event_model.Filler(
        handler_registry, inplace=False, root_map=root_map
    ) as filler:
        with serializer_class(
            directory, file_prefix="documents/{start[uid]}"
        ) as serializer:
            with tqdm(position=0) as progress:
                for name, doc in run.canonical(fill="no"):
                    if external == "fill":
                        name, doc = filler(name, doc)
                        # Omit Resource and Datum[Page] because the data was
                        # filled in place.
                        if name in EXTERNAL_RELATED_DOCS:
                            progress.update()
                            continue
                    elif name == "resource":
                        root = root_map.get(doc["root"], doc["root"])
                        unique_id = root_hash_func(doc["root"])
                        if external is None:
                            if no_documents:
                                root_in_document = doc["root"]
                            else:
                                root_in_document = root
                            # - root_in_document is the 'root' actually in the
                            # resource_document
                            # - root may be different depending on the
                            # source_catalog configuration, which can map the
                            # recorded 'root' in the document to some other
                            # location. This is where we should go looking for
                            # the data if we plan to copy it.
                            # - unique_id is unique to this (root, salt)
                            # combination and used to place the data in a
                            # unique location.
                            key = (root_in_document, root, unique_id)
                            files[key].update(run.get_file_list(doc))
                        if not no_documents:
                            # Replace root with a unique ID before serialization.
                            # We are overriding the local variable name doc here
                            # (yuck!) so that serializer(name, doc) below works on
                            # all document types.
                            doc = doc.copy()
                            doc["root"] = unique_id
                    if not no_documents:
                        serializer(name, doc)
                    progress.update()
    return serializer.artifacts, dict(files)


def write_external_files_manifest(manager, unique_id, files):
    """
    Write a manifest of external files.

    Parameters
    ----------
    manager: suitcase Manager object
    unique_id: Str
    files: Iterable[Union[Str, Path]]
    """
    name = f"external_files_manifest_{unique_id}.txt"
    with manager.open("manifest", name, "xt") as file:
        file.write("\n".join(sorted((str(f) for f in set(files)))))


def copy_external_files(target_directory, root, unique_id, files, strict=False):
    """
    Make a filesystem copy of the external files.

    A filesystem copy is not always applicable/desirable. Use the
    external_file_manifest_*.txt files to feed other file transfer mechanisms,
    such as rsync or globus.

    This is a wrapper around shutil.copyfile.

    Parameters
    ----------
    target_directory: Union[Str, Path]
    root: Str
    files: Iterable[Str]
    strict: Bool, optional
        By default, swallow erros and return a list of them at the end.
        Set to True to debug errors.

    Returns
    -------
    new_root, new_files, failures

    Notes
    -----
    * ``new_root`` is a Path to the new root directory
    * ``new_files`` is the list of filepaths to the files that were created.
    * ``failures`` is a list of uids of runs that raised Exceptions. (The
      relevant tracebacks are logged.)
    """
    new_files = []
    failures = []
    for filename in tqdm(files, total=len(files), desc="Copying external files"):
        relative_path = pathlib.Path(filename).relative_to(root)
        new_root = pathlib.Path(target_directory, unique_id)
        dest = new_root / relative_path
        try:
            os.makedirs(dest.parent, exist_ok=True)
            new_files.append(shutil.copyfile(filename, dest))
        except Exception:
            logger.exception("Error while copying %r to %r", filename, dest)
            if strict:
                raise
            failures.append(filename)
    return new_root, new_files, failures


def write_msgpack_catalog_file(manager, directory, paths, root_map):
    """
    Write a YAML file with configuration for an intake catalog.

    Parameters
    ----------
    manager: suitcase Manager object
    directory: Str
        Directory to which paths below are relative
    paths: Union[Str, List[Str]]
        Relative (s) of JSONL files encoding Documents.
    root_map: Dict
    """
    # Ideally, the drivers should be able to cope with relative paths,
    # interpreting them as relative to the Catalog file. This requires changes
    # to intake (I think) so as a short-term hack, we make the paths aboslute
    # here but note the relative paths in a separate place.
    abs_paths = [str(pathlib.Path(directory, path).absolute()) for path in paths]
    metadata = {
        "generated_by": {
            "library": "databroker_pack",
            "version": get_versions()["version"],
        },
        "relative_paths": [str(path) for path in paths],
    }
    source = {
        "driver": "bluesky-msgpack-catalog",
        "args": {"paths": abs_paths},
        "metadata": metadata,
    }
    if root_map is not None:
        source["args"]["root_map"] = {str(k): str(v) for k, v in root_map.items()}
    sources = {"packed_catalog": source}
    catalog = {"sources": sources}
    FILENAME = "catalog.yml"  # expected by unpack
    with manager.open("catalog_file", FILENAME, "xt") as file:
        yaml.dump(catalog, file)


def write_jsonl_catalog_file(manager, directory, paths, root_map):
    """
    Write a YAML file with configuration for an intake catalog.

    Parameters
    ----------
    manager: suitcase Manager object
    directory: Str
        Directory to which paths below are relative
    paths: Union[Str, List[Str]]
        Relative (s) of JSONL files encoding Documents.
    root_map: Dict
    """
    # There is clearly some code repetition here with respect to
    # write_msgpack_catalog_file, but I expect they may diverge over time as
    # the suitcase implementation pick up format-specific options.

    # Ideally, the drivers should be able to cope with relative paths,
    # interpreting them as relative to the Catalog file. This requires changes
    # to intake (I think) so as a short-term hack, we make the paths aboslute
    # here but note the relative paths in a separate place.
    abs_paths = [str(pathlib.Path(directory, path).absolute()) for path in paths]
    metadata = {
        "generated_by": {
            "library": "databroker_pack",
            "version": get_versions()["version"],
        },
        "relative_paths": [str(path) for path in paths],
    }
    source = {
        "driver": "bluesky-jsonl-catalog",
        "args": {"paths": abs_paths},
        "metadata": metadata,
    }
    if root_map is not None:
        source["args"]["root_map"] = {str(k): str(v) for k, v in root_map.items()}
    sources = {"packed_catalog": source}
    catalog = {"sources": sources}
    FILENAME = "catalog.yml"
    with manager.open("catalog_file", FILENAME, "xt") as file:
        yaml.dump(catalog, file)


def write_documents_manifest(manager, directory, artifacts):
    """
    Wirte the paths to all the files of Documents relative to the pack directory.

    Parameters
    ----------
    manager: suitcase Manager object
    directory: Str
        Pack directory
    artifacts: List[Str]
    """
    FILENAME = "documents_manifest.txt"
    abs_directory = pathlib.Path(directory).absolute()
    with manager.open("documents_manifest", FILENAME, "xt") as file:
        for artifact in artifacts:
            file.write(f"{pathlib.Path(artifact).relative_to(abs_directory)!s}\n")



import os
import pathlib

import pytest

from .._pack import copy_external_files


def test_strict_copy_external_files(tmpdir):
    root = str(pathlib.Path(tmpdir, "source"))
    os.makedirs(root)
    target_directory = str(pathlib.Path(tmpdir, "dest"))
    filepath = pathlib.Path(root, "testfile")
    with open(filepath, "w") as file:
        file.write("placeholder")

    # Test a successful copy.
    unique_id = "placeholder_unique_id"
    _, _, failures = copy_external_files(
        target_directory, root, unique_id, [pathlib.Path(root, "testfile")], strict=True
    )
    assert not failures
    with open(pathlib.Path(target_directory, unique_id, "testfile")) as file:
        assert file.read() == "placeholder"

    # Test failures with and without strict.
    nonexistant_file = pathlib.Path(root, unique_id, "DOES_NOT_EXIST")
    with pytest.raises(FileNotFoundError):
        copy_external_files(
            target_directory, root, unique_id, [nonexistant_file], strict=True,
        )
    _, _, failures = copy_external_files(
        target_directory, root, unique_id, [nonexistant_file], strict=False,
    )
    assert failures == [nonexistant_file]



#!/usr/bin/env python
import argparse
import functools
import logging
import os
import pathlib
import sys
import tempfile

from suitcase.utils import MultiFileManager

from ._utils import ListCatalogsAction, ShowVersionAction
from .._pack import (
    copy_external_files,
    export_catalog,
    export_uids,
    write_documents_manifest,
    write_jsonl_catalog_file,
    write_msgpack_catalog_file,
    write_external_files_manifest,
)


print = functools.partial(print, file=sys.stderr)


def main():
    with tempfile.NamedTemporaryFile("w", delete=False) as file:
        error_logfile_name = file.name
    error_handler = logging.FileHandler(error_logfile_name)
    logging.getLogger("databroker_pack").addHandler(error_handler)
    parser = argparse.ArgumentParser(
        description="Pack up some Bluesky Runs into portable files.",
        formatter_class=argparse.RawTextHelpFormatter,
        epilog="""
*~ Examples ~*

List the available options for CATALOG and exit.

$ databroker-pack --list-catalogs
...

Export Runs from a range of time.

$ databroker-pack CATALOG -q "TimeRange(since='2020')" DIRECTORY
$ databroker-pack CATALOG -q "TimeRange(since='2020', until='2020-03-01)" DIRECTORY

Export Runs from a range of time with a certain plan_name.

$ databroker-pack CATALOG -q "TimeRange(since='2020')" -q "{'plan_name': 'count'}" DIRECTORY

Export a specific Run by its scan_id

$ databroker-pack CATALOG -q "{'scan_id': 126360}" DIRECTORY

Export specific Runs given by their Run Start UID (or the first several
characters) entered at the command prompt...

$ databroker-pack CATALOG --uids -
3c93c54e
47587fa8
ebad8c01
<Ctrl D>

...or read from a file.

$ databroker-pack CATALOG --uids uids_to_pack.txt

Export an entire catalog.

$ databroker-pack CATALOG --all DIRECTORY

Copy the external data files into the output directory.

$ databroker-pack CATALOG --all --copy-external DIRECTORY
""",
    )
    parser.register("action", "list_catalogs", ListCatalogsAction)
    parser.register("action", "show_version", ShowVersionAction)
    filter_group = parser.add_argument_group(
        description="Which Runs should we pack? Must specify one of these:"
    ).add_mutually_exclusive_group()
    external_group = parser.add_argument_group(
        description=(
            "What should we do with external files (e.g. large array data "
            "files written by detectors)? By default, we will write a text "
            "file with all the relevant file paths. Other options:"
        )
    ).add_mutually_exclusive_group()
    # By making a group "other" we can place the less-frequently-used options
    # at the bottom.
    other_group = parser.add_argument_group()
    parser.add_argument("catalog", type=str, help="Catalog name")
    parser.add_argument("directory", type=str, help="Path to output directory")
    parser.add_argument(
        "--list-catalogs",
        action="list_catalogs",
        default=argparse.SUPPRESS,
        help="List allowed values for catalog and exit.",
    )
    filter_group.add_argument(
        "--all", action="store_true", help="Export every Run in this Catalog."
    )
    filter_group.add_argument(
        "-q",
        "--query",
        type=str,
        action="append",
        help=(
            "MongoDB-style query or databroker.queries Query. "
            "Narrow results by chaining multiple queries like "
            "-q \"TimeRange(since='2020')\" "
            "-q \"{'sample': 'Au'}\""
        ),
    )
    filter_group.add_argument(
        "--uids",
        type=argparse.FileType("r"),
        action="append",
        help=("Newline-separated (partial) uids. Lines starting with # are skipped."),
    )
    external_group.add_argument(
        "--copy-external",
        action="store_true",
        help="Copy relevant external files into the output directory.",
    )
    external_group.add_argument(
        "--fill-external",
        action="store_true",
        help="Place external data directly in the documents.",
    )
    external_group.add_argument(
        "--ignore-external",
        action="store_true",
        help=(
            "By default, the locations of all relevant external files on the "
            "source machine are written to text files. "
            "Set this to omit those manifests."
        ),
    )
    other_group.add_argument(
        "--format",
        default="msgpack",
        choices=["msgpack", "jsonl"],
        type=str.lower,  # Makes choices case-insensitive
        help=(
            "Format for Documents in the pack output. Choise msgpack "
            "(default) for binary speed/compactness. Choose jsonl for "
            "plaintext."
        ),
    )
    other_group.add_argument(
        "--no-documents", action="store_true", help="Do not pack the Documents.",
    )
    other_group.add_argument(
        "--limit",
        type=int,
        help=(
            "Stop after exporting some number of Runs. Useful for testing a "
            "subset before doing a lengthy export."
        ),
    )
    other_group.add_argument(
        "--strict",
        action="store_true",
        help=(
            "Exit when error occurs. Otherwise failures are logged to "
            "stderr as they happen "
            "and again all together at the end."
        ),
    )
    other_group.add_argument(
        "--handler-registry",
        help=(
            "Dict mapping specs to handler objects like "
            "\"{'AD_HDF5': 'area_detector_handlers.handlers:AreaDetectorHDF5Handler'}\" "
            "If unspecified, automatic handler discovery is used."
        ),
    )
    other_group.add_argument(
        "--salt",
        type=str.encode,  # casts input to bytes
        help=("Set this to override the random default with a fixed value."),
    )
    other_group.add_argument(
        "-V",
        "--version",
        action="show_version",
        default=argparse.SUPPRESS,
        help="Show databroker_pack version and exit.",
    )
    args = parser.parse_args()
    # We hide the imports here just for speed.
    if args.format == "msgpack":
        import suitcase.msgpack

        serializer_class = suitcase.msgpack.Serializer
    if args.format == "jsonl":
        import suitcase.jsonl

        serializer_class = suitcase.jsonl.Serializer
    if args.ignore_external:
        external = "ignore"
    elif args.fill_external:
        external = "fill"
    else:
        external = None
    import databroker
    import databroker.queries
    import databroker.core

    if args.handler_registry is not None:
        # Temporary limitation due to the fact that BlueskyRun.get_file_list()
        # does not accept a custom Filler.
        if not args.fill_external:
            raise NotImplementedError(
                "The --handler-registry parameter currently only works with "
                "--fill-external. All other modes of handling external data "
                "use automatic handler discovery only."
            )
        # There are two kinds of "parsing" happening here.
        # 1. Go from stdin string to dict using ast.literal_eval.
        # 2. Go from dict of strings to dict of actual handler classes using
        #    parse_handler_registry.
        try:
            import ast

            handler_registry = ast.literal_eval(args.handler_registry)
        except Exception as exc:
            raise ValueError(
                f"Could not parse --handler-registry {args.handler_registry}. "
                "A dict of strings is expected."
            ) from exc
        handler_registry = databroker.core.parse_handler_registry(handler_registry)
    else:
        handler_registry = None
    # The MultiFileManager would create this directory for us on demand (when
    # we open the first file), but let's create it early so we can check
    # permissions.
    try:
        os.makedirs(args.directory, exist_ok=True)
    except Exception as exc:
        print(f"Could not create directory at {args.directory}.")
        print(f"Error: {exc!r}")
        sys.exit(1)
    if not os.access(args.directory, os.W_OK):
        print(f"Directory at {args.directory} is not writable.")
        sys.exit(1)
    catalog = databroker.catalog[args.catalog]()
    manager = MultiFileManager(args.directory)
    try:
        # We write the catalog file at the end because we need to collect
        # information to write the root_map. We never overwrite existing files,
        # so we will fail if catalog.yml exists already. In order to save time
        # and fail early, check that the file exists.
        catalog_file_path = pathlib.Path(args.directory, "catalog.yml")
        if os.path.isfile(catalog_file_path):
            print(
                f"The file {catalog_file_path} exists. Specify an empty directory, or "
                "a nonexistent one (which will be created by "
                f"{os.path.basename(sys.argv[0])})."
            )
            sys.exit(1)
        if args.query or args.all:
            if args.all:
                # --all is an alias for --query "{}"
                raw_queries = ["{}"]
            else:
                raw_queries = args.query
            queries = []
            for raw_query in raw_queries:
                # Parse string like "{'scan_id': 123}" to dict.
                try:
                    query = dict(eval(raw_query, vars(databroker.queries)))
                except Exception:
                    raise ValueError(f"Could not parse query {raw_query}.")
                queries.append(query)
            combined_query = {"$and": queries}
            # HACK We need no_cursor_timeout only until databroker learns to
            # seamlessly remake cursors when they time out.
            try:
                results = catalog.search(combined_query, no_cursor_timeout=True)
            except TypeError:
                # Drivers that are not Mongo drivers will not support
                # no_cursor_timeout.
                results = catalog.search(combined_query)
            if not results:
                print(f"Query {combined_query} yielded no results. Exiting.")
                sys.exit(1)
            artifacts, external_files, failures, file_uids = export_catalog(
                results,
                manager,
                strict=args.strict,
                external=external,
                no_documents=args.no_documents,
                handler_registry=handler_registry,
                serializer_class=serializer_class,
                salt=args.salt,
                limit=args.limit,
            )
        elif args.uids:
            # Skip blank lines and commented lines.
            uids = []
            for uid_file in args.uids:
                uids.extend(
                    line.strip()
                    for line in uid_file.read().splitlines()
                    if line and not line.startswith("#")
                )
            if not uids:
                print("Found empty input for --uids. Exiting")
                sys.exit(1)
            artifacts, external_files, failures, file_uids = export_uids(
                catalog,
                uids[:args.limit],
                manager,
                strict=args.strict,
                external=external,
                no_documents=args.no_documents,
                handler_registry=handler_registry,
                serializer_class=serializer_class,
                salt=args.salt,
            )
        else:
            parser.error(
                "Must specify which Runs to pack, --query ... or "
                "--uids ... or --all."
            )
        if not args.no_documents and artifacts.get("all"):
            write_documents_manifest(manager, args.directory, artifacts["all"])
        root_map = {}
        copying_failures = []
        if external is None:
            # When external is None, external data is neither being filled into
            # the Documents (external == 'fill') nor ignored (external ==
            # 'ignore') so we have to provide a root_map in the catalog file to
            # reference its location.
            if args.copy_external:
                target_drectory = pathlib.Path(args.directory, "external_files")
                for (
                    (root_in_document, root, unique_id),
                    files,
                ) in external_files.items():
                    new_root, new_files, copying_failures_ = copy_external_files(
                        target_drectory, root, unique_id, files, strict=args.strict
                    )
                    copying_failures.extend(copying_failures_)
                    # The root_map value will be the relative path to
                    # the data within the pack directory.
                    relative_root = new_root.relative_to(args.directory)
                    if not args.no_documents:
                        # When we are exporting documents, we rewrite the
                        # 'root' key in the Resource to unique_id to ensure
                        # no collisions of the keys in root_map.
                        root_map.update({unique_id: relative_root})
                    else:
                        # If we are not exporting documents, the root_map has
                        # to refer to the root as it is.
                        root_map.update({root_in_document: relative_root})
                    rel_paths = [
                        pathlib.Path(f).relative_to(args.directory) for f in new_files
                    ]
                    write_external_files_manifest(manager, unique_id, rel_paths)
            else:
                for (
                    (root_in_document, root, unique_id),
                    files,
                ) in external_files.items():
                    # The root_map value will be the current absolute path to
                    # the data.
                    if not args.no_documents:
                        # When we are exporting documents, we rewrite the
                        # 'root' key in the Resource to unique_id to ensure
                        # no collisions of the keys in root_map.
                        root_map.update({unique_id: root})
                    else:
                        # If we are not exporting documents, the root_map has
                        # to refer to the root as it is.
                        root_map.update({root_in_document: root})
                    write_external_files_manifest(manager, unique_id, files)
        if args.format == "jsonl":
            paths = ["./documents/*.jsonl"]
            write_jsonl_catalog_file(manager, args.directory, paths, root_map)
        elif args.format == "msgpack":
            paths = ["./documents/*.msgpack"]
            write_msgpack_catalog_file(manager, args.directory, paths, root_map)
        # No need for an else here; we validated that it is one of these above.
        if failures:
            print(f"{len(failures)} Runs failed to pack.")
            with tempfile.NamedTemporaryFile("w", delete=False) as file:
                print(f"See {file.name} for a list of uids of Runs that failed.")
                file.write("\n".join(failures))
        if copying_failures:
            with tempfile.NamedTemporaryFile("w", delete=False) as file:
                print(f"See {file.name} for a list of files that failed to copy.")
                file.write("\n".join(copying_failures))
        if failures or copying_failures:
            print(f"See {error_logfile_name} for error logs with more information.")
            sys.exit(1)
    finally:
        manager.close()


if __name__ == "__main__":
    main()



import copy
import os
import pathlib

import databroker
import intake
from tqdm import tqdm
import yaml

from ._utils import CatalogNameExists

__all__ = ("unpack_inplace", "unpack_mongo_normalized")


def unpack_inplace(path, catalog_name, merge=False):
    """
    Place a catalog configuration file in the user configuration area.

    Parameters
    ----------
    path: Path
        Path to output from pack
    catalog_name: Str
        A unique name for the catalog
    merge: Boolean, optional
        Unpack into an existing catalog

    Returns
    -------
    config_path: Path
        Location of new catalog configuration file
    """
    # Validate user input.
    if not os.path.isdir(path):
        raise ValueError(f"{path} is not a directory")
    source_catalog_file_path = pathlib.Path(path, "catalog.yml")
    if not os.path.isfile(source_catalog_file_path):
        raise ValueError(f"Could not find 'catalog.yml' in {path}")
    dest_catalog_file_name = f"databroker_unpack_{catalog_name}.yml"
    config_dir = databroker.catalog_search_path()[0]
    dest_catalog_file_path = pathlib.Path(config_dir, dest_catalog_file_name)
    exists = catalog_name in databroker.catalog
    if exists:
        if not merge:
            raise CatalogNameExists(catalog_name)
        if not os.path.isfile(dest_catalog_file_path):
            raise ValueError(
                "The catalog exists but not in the user-writable location. "
                "Pick a different catalog name."
            )
        with open(dest_catalog_file_path) as file:
            existing_catalog = yaml.safe_load(file)
        existing_source = existing_catalog["sources"][catalog_name]
    else:
        existing_source = None

    with open(source_catalog_file_path) as file:
        catalog = yaml.safe_load(file)
    source = catalog["sources"].pop("packed_catalog")

    # Check that the drivers match. For example, we cannot merge a
    # msgpack-backed catalog into an existing JSONL-backed one.
    if existing_source is not None:
        source_driver = source.get("driver")
        existing_source_driver = existing_source.get("driver")
        if existing_source_driver != source_driver:
            raise ValueError(
                f"Cannot merge source with driver {source_driver} into source "
                f"with driver {existing_source_driver}"
            )

    # Handle temporary condition where 'pack' puts absolute paths in "args"
    # and puts relative paths off to the side.
    relative_paths = source.get("metadata", {}).get("relative_paths")
    if relative_paths:
        new_paths = [
            str(pathlib.Path(path, rel_path).absolute()) for rel_path in relative_paths
        ]
        source["args"]["paths"] = sorted(new_paths)

    # Merge paths and relative_paths from existing source, if applicable.
    if existing_source is not None:
        source["args"]["paths"] = sorted(
            set(source["args"]["paths"]) | set(existing_source["args"]["paths"])
        )
        if "metadata" not in source:
            source["metadata"] = {}
        source["metadata"]["relative_paths"] = sorted(
            set(source["metadata"].get("relative_paths", []))
            | set(existing_source.get("metadata", {}).get("relative_paths", []))
        )

    # The root_map values may be relative inside a pack, given relative to the
    # catalog file. Now that we are going to use a catalog file in a config
    # directory, we need to make these paths absolute.
    for k, v in source["args"].get("root_map", {}).items():
        if not pathlib.Path(v).is_absolute():
            source["args"]["root_map"][k] = str(pathlib.Path(path, v).absolute())

    # Merge root_map from existing source, if applicable.
    if existing_source is not None:
        collisions = set(source["args"].get("root_map", {})).intersection(
            set(existing_source["args"].get("root_map", {}))
        )
        if collisions:
            raise ValueError(
                "root_map between existing source and new source have "
                f"colliding keys {collisions}"
            )
        source["args"]["root_map"].update(existing_source["args"].get("root_map", {}))

    catalog["sources"][catalog_name] = source
    os.makedirs(config_dir, exist_ok=True)
    with open(dest_catalog_file_path, "w") as file:
        yaml.dump(catalog, file)
    return dest_catalog_file_path


def unpack_mongo_normalized(path, uri, catalog_name, merge=False):
    """
    Place a catalog configuration file in the user configuration area.

    Parameters
    ----------
    path: Path
        Path to output from pack
    uri: Str
        MongoDB URI. Must include a database name. Example:
        ``mongodb://localhost:27017/databroker_unpack_my_catalog``
    catalog_name: Str
        A unique name for the catalog
    merge: Boolean, optional
        Unpack into an existing catalog

    Returns
    -------
    config_path: Path
        Location of new catalog configuration file
    """
    import pymongo
    import suitcase.mongo_normalized

    # Validate user input.
    if not os.path.isdir(path):
        raise ValueError(f"{path} is not a directory")
    source_catalog_file_path = pathlib.Path(path, "catalog.yml")
    if not os.path.isfile(source_catalog_file_path):
        raise ValueError(f"Could not find 'catalog.yml' in {path}")
    dest_catalog_file_name = f"databroker_unpack_{catalog_name}.yml"
    config_dir = databroker.catalog_search_path()[0]
    dest_catalog_file_path = pathlib.Path(config_dir, dest_catalog_file_name)
    exists = catalog_name in databroker.catalog
    if exists:
        if not merge:
            raise CatalogNameExists(catalog_name)
        if not os.path.isfile(dest_catalog_file_path):
            raise ValueError(
                "The catalog exists but not in the user-writable location. "
                "Pick a different catalog name."
            )
        with open(dest_catalog_file_path) as file:
            existing_catalog = yaml.safe_load(file)
        existing_source = existing_catalog["sources"][catalog_name]
    else:
        existing_source = None

    with open(source_catalog_file_path) as file:
        catalog = yaml.safe_load(file)
    source = catalog["sources"].pop("packed_catalog")

    # Ensure that the URI has a database name in it.
    if not pymongo.uri_parser.parse_uri(uri)["database"]:
        raise ValueError("Mongo URI must include a database name.")
    database = pymongo.MongoClient(uri).get_database()

    # Check that the target catalog is the right driver and URIs.
    if existing_source is not None:
        existing_source_driver = existing_source.get("driver")
        if existing_source_driver != "bluesky-mongo-normalized-catalog":
            raise ValueError(
                f"Existing catalog has driver {existing_source_driver} "
                "so we cannot make a MongoDB-backed catalog with that name."
            )
        if existing_source["args"]["metadatastore_db"] != uri:
            raise ValueError(
                "Existing catalog has metadatastore_db"
                f"{existing_source['args']['metadatastore_db']} "
                r"which does not match requested uri {uri}."
            )
        if existing_source["args"]["asset_registry_db"] != uri:
            raise ValueError(
                "Existing catalog has asset_registry_db"
                f"{existing_source['args']['asset_registry_db']} "
                r"which does not match requested uri {uri}."
            )

    # Handle temporary condition where 'pack' puts absolute paths in "args"
    # and puts relative paths off to the side.
    relative_paths = source.get("metadata", {}).get("relative_paths")
    if relative_paths:
        new_paths = [
            str(pathlib.Path(path, rel_path).absolute()) for rel_path in relative_paths
        ]
        source["args"]["paths"] = sorted(new_paths)

    # The root_map values may be relative inside a pack, given relative to the
    # catalog file. Now that we are going to use a catalog file in a config
    # directory, we need to make these paths absolute.
    for k, v in source["args"].get("root_map", {}).items():
        if not pathlib.Path(v).is_absolute():
            source["args"]["root_map"][k] = str(pathlib.Path(path, v).absolute())

    # Copy data into MongoDB.
    catalog_class = intake.registry[source["driver"]]
    source_catalog = catalog_class(**source["args"])
    serializer = suitcase.mongo_normalized.Serializer(
        metadatastore_db=database, asset_registry_db=database,
    )
    with tqdm(
        total=len(source_catalog), desc="Copying Documents into MongoDB"
    ) as progress:
        for uid, run in source_catalog.items():
            for name, doc in run.canonical(fill="no"):
                serializer(name, doc)
            progress.update()

    # Modify a copy of the original, file-based source, to make it a
    # mongo_normalized one.
    mongo_source = copy.deepcopy(source)
    mongo_source["driver"] = "bluesky-mongo-normalized-catalog"
    mongo_source["args"] = {
        "metadatastore_db": uri,
        "asset_registry_db": uri,
        "root_map": source["args"]["root_map"],
    }

    # Merge root_map from existing source, if applicable.
    if existing_source is not None:
        collisions = set(mongo_source["args"].get("root_map", {})).intersection(
            set(existing_source["args"].get("root_map", {}))
        )
        if collisions:
            raise ValueError(
                "root_map between existing source and new source have "
                f"colliding keys {collisions}"
            )
        mongo_source["args"]["root_map"].update(
            existing_source["args"].get("root_map", {})
        )

    catalog["sources"][catalog_name] = mongo_source
    os.makedirs(config_dir, exist_ok=True)
    with open(dest_catalog_file_path, "w") as file:
        yaml.dump(catalog, file)
    return dest_catalog_file_path



import pathlib
import subprocess
import sys

import pytest
import yaml


@pytest.mark.parametrize(
    "cli_args",
    [
        ["--version"],
        ["--list-catalogs"],
        ["--all"],
        ["--query", "{}"],
        ["--all", "--no-documents"],
        ["--all", "--strict"],
        ["--all", "--limit", "1"],
        ["--all", "--salt", "asdf"],
        ["--all", "--ignore-external"],
        ["--all", "--copy-external"],
        ["--all", "--fill-external"],
        [
            "--all",
            "--fill-external",
            "--handler-registry",
            "{'NPY_SEQ': 'ophyd.sim.NumpySeqHandler'}",
        ],
        # Repeat some of the above with JSONL instead of the default msgpack.
        # We could use a second layer of parametrize here but that seems more
        # confusing than helpful.
        ["--format", "jsonl", "--query", "{}"],
        ["--format", "jsonl", "--all", "--no-documents"],
        ["--format", "jsonl", "--all", "--strict"],
        ["--format", "jsonl", "--all", "--limit", "1"],
        ["--format", "jsonl", "--all", "--salt", "asdf"],
        ["--format", "jsonl", "--all", "--ignore-external"],
        ["--format", "jsonl", "--all", "--copy-external"],
        ["--format", "jsonl", "--all", "--fill-external"],
        [
            "--format",
            "jsonl",
            "--all",
            "--fill-external",
            "--handler-registry",
            "{'NPY_SEQ': 'ophyd.sim.NumpySeqHandler'}",
        ],
    ],
)
@pytest.mark.parametrize("relative_target_directory", [True, False])
def test_pack_smoke(cli_args, simple_catalog, tmpdir, relative_target_directory):
    "Smoke test common options."
    TIMEOUT = 10
    CATALOG = simple_catalog
    if relative_target_directory:
        p = subprocess.Popen(
            [
                sys.executable,
                "-um",
                "databroker_pack.commandline.pack",
                CATALOG,
                pathlib.Path(tmpdir).parts[-1],
                *cli_args,
            ],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            cwd=str(pathlib.Path(tmpdir).parent),
        )
    else:
        p = subprocess.Popen(
            [
                sys.executable,
                "-um",
                "databroker_pack.commandline.pack",
                CATALOG,
                tmpdir,
                *cli_args,
            ],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
    # Capture stdout, stderr for interactive debugging.
    stdout, stderr = p.communicate(TIMEOUT)
    assert p.returncode == 0


HASH = "3e3b48cf1579c708c4b515de4927d03a"  # stable because we give --salt SALT
REL_PATH = str(pathlib.Path("external_files", HASH))
ORIGINAL_ROOT = "f8483b80395561ac779a5425da6646ca"
ABS_PATH = str(
    pathlib.Path(__file__).parent
    / "data"
    / "simple_catalog"
    / "external_files"
    / ORIGINAL_ROOT
)


@pytest.mark.parametrize(
    "cli_args, expected",
    [
        ([], {HASH: ABS_PATH}),
        (["--no-documents"], {ORIGINAL_ROOT: ABS_PATH}),
        (["--copy-external"], {HASH: REL_PATH}),
        (["--no-documents", "--copy-external"], {ORIGINAL_ROOT: REL_PATH}),
    ],
    ids=[
        "hash_to_abs_path",
        "original_root_to_abs_path",
        "hash_to_rel_path",
        "original_root_to_rel_path",
    ],
)
@pytest.mark.parametrize("relative_target_directory", [True, False])
def test_root_map(
    cli_args, expected, simple_catalog, tmpdir, relative_target_directory
):
    "Smoke test common options."
    TIMEOUT = 10
    CATALOG = simple_catalog
    if relative_target_directory:
        p = subprocess.Popen(
            [
                sys.executable,
                "-um",
                "databroker_pack.commandline.pack",
                CATALOG,
                pathlib.Path(tmpdir).parts[-1],
                "--all",
                "--salt",
                "SALT",
                *cli_args,
            ],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            cwd=str(pathlib.Path(tmpdir).parent),
        )
    else:
        p = subprocess.Popen(
            [
                sys.executable,
                "-um",
                "databroker_pack.commandline.pack",
                CATALOG,
                tmpdir,
                "--all",
                "--salt",
                "SALT",
                *cli_args,
            ],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
    # Capture stdout, stderr for interactive debugging.
    stdout, stderr = p.communicate(TIMEOUT)
    assert p.returncode == 0
    with open(pathlib.Path(tmpdir, "catalog.yml")) as file:
        catalog = yaml.safe_load(file)
    root_map = catalog["sources"]["packed_catalog"]["args"]["root_map"]
    assert root_map == expected



#!/usr/bin/env python
import argparse
import sys
from .._unpack import unpack_inplace, unpack_mongo_normalized
from ._utils import ListCatalogsAction, ShowVersionAction
from .._utils import CatalogNameExists


def main():
    parser = argparse.ArgumentParser(
        description="Install a Catalog of packed Bluesky Runs.",
        formatter_class=argparse.RawTextHelpFormatter,
    )
    parser.register("action", "show_version", ShowVersionAction)
    parser.register("action", "list_catalogs", ListCatalogsAction)
    parser.add_argument(
        "how",
        type=str,
        choices=("inplace", "mongo_normalized"),
        help="Read Documents in place (from files) or load them into a database.",
    )
    parser.add_argument("path", type=str, help="Path to pack directory")
    parser.add_argument("name", type=str, help="Name of new catalog")
    parser.add_argument("--no-merge", action="store_true")
    parser.add_argument(
        "--mongo-uri",
        type=str,
        help=(
            "MongoDB URI. Default is 'mongodb://localhost:27017/{database}' "
            "where the token {database}, if given, is filled in with "
            "'databroker_{name}. The value of {name} is taken from the name "
            "parameter above."
        ),
    )
    parser.add_argument(
        "--list-catalogs",
        action="list_catalogs",
        default=argparse.SUPPRESS,
        help="List allowed values for catalog and exit.",
    )
    parser.add_argument(
        "-V",
        "--version",
        action="show_version",
        default=argparse.SUPPRESS,
        help="Show databroker_pack version and exit.",
    )
    args = parser.parse_args()
    try:
        if args.how == "inplace":
            config_path = unpack_inplace(args.path, args.name, merge=not args.no_merge)
        elif args.how == "mongo_normalized":
            uri = args.mongo_uri or "mongodb://localhost:27017/{database}"
            formatted_uri = uri.format(database=f"databroker_{args.name}")
            config_path = unpack_mongo_normalized(
                args.path, formatted_uri, args.name, merge=not args.no_merge
            )
        # We rely on argparse to ensure that args.how is one of the above.
    except CatalogNameExists:
        import databroker
        import itertools
        import shutil

        term_size = shutil.get_terminal_size((100, 50))
        extant_catalogs = sorted(databroker.catalog)
        ncats = len(extant_catalogs)
        col_width = max(len(_) for _ in extant_catalogs) + 5
        ncols = max(term_size.columns // col_width, 1)

        format_str = f"{{:<{col_width}}}" * ncols

        n_rows = (ncats // ncols) + int(ncols % ncols > 0)

        cols = [extant_catalogs[j * n_rows: (j + 1) * n_rows] for j in range(ncols)]

        nice_names = "\n".join(
            format_str.format(*g) for g in itertools.zip_longest(*cols, fillvalue="")
        )

        msg = f"""
You tried to unpack to a catalog named {args.name} which already exists.

The currently existing catalogs on your system are:

{nice_names}

Please try unpacking again with a new name.
"""
        print(msg)
        sys.exit(1)
    else:
        print(f"Placed configuration file at {config_path!s}")


if __name__ == "__main__":
    main()



import os
from pathlib import PurePath
from pkg_resources import resource_filename

import databroker
import pytest

from databroker_pack._unpack import unpack_inplace

DATA_DIR = PurePath(resource_filename('databroker_pack', 'tests/data'))


@pytest.fixture(scope="module")
def simple_catalog():
    DIRECTORY = "simple_catalog"
    NAME = "databroker_pack_tests_simple_catalog"
    config_path = unpack_inplace(str(DATA_DIR.joinpath(DIRECTORY)), NAME)
    try:
        # Use this once list_configs() is fixed to be complete.
        # It is incomplete in databroker 1.0.0.
        # assert NAME in databroker.utils.list_configs()
        databroker.catalog.force_reload()
        assert NAME in list(databroker.catalog)
        yield NAME
    finally:
        os.unlink(config_path)



import argparse
from .._version import get_versions


class ShowVersionAction(argparse.Action):
    # a special action that allows the usage --version to override
    # any 'required args' requirements, the same way that --help does

    def __init__(
        self,
        option_strings,
        dest=argparse.SUPPRESS,
        default=argparse.SUPPRESS,
        help=None,
    ):
        super().__init__(
            option_strings=option_strings,
            dest=dest,
            default=default,
            nargs=0,
            help=help,
        )

    def __call__(self, parser, namespace, values, option_string=None):
        print(get_versions()["version"])
        parser.exit()


class ListCatalogsAction(argparse.Action):
    # a special action that allows the usage --list-catalogs to override
    # any 'required args' requirements, the same way that --help does

    def __init__(
        self,
        option_strings,
        dest=argparse.SUPPRESS,
        default=argparse.SUPPRESS,
        help=None,
    ):
        super().__init__(
            option_strings=option_strings,
            dest=dest,
            default=default,
            nargs=0,
            help=help,
        )

    def __call__(self, parser, namespace, values, option_string=None):
        import databroker

        for name in databroker.catalog:
            print(name)
        parser.exit()
